# -*- coding: utf-8 -*-
"""
    Copyright (C) 2021  Opsdis AB

    This file is part of indis - Icinga native directory importer service.

    indis is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    indis is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with indis.  If not, see <http://www.gnu.org/licenses/>.

"""

from indis.model.dependency import Dependency


class HostDependency(Dependency):

    def __init__(self, name: str, parent_host: str, child_host: str):
        super().__init__(name=name, parent_host=parent_host, child_host=child_host)



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2021  Opsdis AB

    This file is part of indis - Icinga native directory importer service.

    indis is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    indis is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with indis.  If not, see <http://www.gnu.org/licenses/>.

"""

import os

import yaml


class Singleton(type):
    """
    Provide singleton pattern to MonitorConfig. A new instance is only created if:
     - instance do not exists
     - config is provide in constructor call, __init__
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances or args:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Configuration(object, metaclass=Singleton):

    def __init__(self, configfile=None):
        self.conf = {}
        try:
            if os.getenv('IND_CONF_FILE'):
                with open(os.getenv('IND_CONF_FILE'), 'r', encoding='utf-8') as yml_file:
                    self.conf = yaml.load(yml_file, Loader=yaml.FullLoader)

            elif configfile is None:
                with open("config.yml", 'r', encoding='utf-8') as yml_file:
                    self.conf = yaml.load(yml_file, Loader=yaml.FullLoader)
            else:
                with open(configfile, 'r', encoding='utf-8') as yml_file:
                    self.conf = yaml.load(yml_file, Loader=yaml.FullLoader)
        except FileNotFoundError as err:
            raise err

    @staticmethod
    def get_keys():
        return Configuration().conf.keys()

    @staticmethod
    def get(name):
        if name in Configuration().conf:
            return Configuration().conf[name]
        else:
            return None



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2021  Opsdis AB

    This file is part of indis - Icinga native directory importer service.

    indis is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    indis is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with indis.  If not, see <http://www.gnu.org/licenses/>.

"""
import json
from typing import List, Dict, Any

from indis.model.common import Common


class Zone(Common):
    """
    Command is the same for CheckCommand, NotificationCommand and EventCommand
    """
    __initialized = False

    def __init__(self, name: str, object_type: str = 'object'):
        super().__init__(name=name, object_type=object_type)
        # Array of object names	Optional. Array of endpoint names located in this zone.
        self.endpoints: List[str] = list()
        # Object name	Optional. The name of the parent zone. (Do not specify a global zone)
        self.parent: str = ''
        # Boolean	Optional. Whether configuration files for this zone should be synced to all endpoints. Defaults to false.
        # in icinga this is named global, but since its a reserved word in python its called is_global.
        # This is translated when json is generated
        self.is_global: bool = False

        self._ind = {'object_name', 'object_type'}
        self.__initialized = True

    def __setattr__(self, name, value):
        if self.__initialized:

            self._ind.add(name)
            object.__setattr__(self, name, value)
        else:
            object.__setattr__(self, name, value)

    # "Override" common since must handle translation from is_global to global
    def to_json(self) -> str:
        res = self.to_dict()
        return json.dumps(res)

    # "Override" common since must handle translation from is_global to global
    def to_dict(self) -> Dict[str, Any]:
        res = dict()
        for key, value in self.__dict__.items():
            if key.startswith('_'):
                # Remove all protected attributes
                continue
            if key in self._ind:
                if key == 'is_global':
                    res['global'] = value
                else:
                    res[key] = value
            elif isinstance(self.__dict__[key], dict) and self.__dict__[key]:
                res[key] = value
            elif isinstance(self.__dict__[key], list) and self.__dict__[key]:
                res[key] = value
            elif key == 'is_global':
                res['global'] = value
        return res



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2021  Opsdis AB

    This file is part of indis - Icinga native directory importer service.

    indis is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    indis is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with indis.  If not, see <http://www.gnu.org/licenses/>.

"""

from indis.model.group import Group
from indis.processor.common import Processing


class Groups(Processing):

    def process(self) -> int:
        pc = self.config
        count = 0
        if pc and 'groups' in pc:

            create_hostgroups = set()
            for object_type in self.transfer.get_keys():
                if object_type in pc.get('groups'):
                    # For each object specific
                    for group in pc.get('groups')[object_type]:
                        for object_name in self.transfer.hosts.keys():
                            self.transfer.hosts[object_name].groups.append(group)
                            create_hostgroups.add(Group(name=group))
                            count += 1

            # Add hostgroups that is added since it may not exists
            for host_group in create_hostgroups:
                self.transfer.hostgroups[host_group.object_name] = host_group

        return count



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2021  Opsdis AB

    This file is part of indis - Icinga native directory importer service.

    indis is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    indis is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with indis.  If not, see <http://www.gnu.org/licenses/>.

"""

import datetime
import json
import logging
import numbers
import os
import sys
from typing import Any

from dateutil.tz import tzutc

from indis.configuration import Configuration as conf

MESSAGE = 'message'


class Iso8601UTCTimeFormatter(logging.Formatter):

    def __init__(self, log_format=None, time_format=None):
        """
        The purpose of this constructor is to set the timezone to UTC for later use.

        :param log_format: Log record formatting string.
        :type log_format: str
        :param time_format: Time formatting string. You probably **DO NOT** want one.
        :type time_format: str
        """
        super(Iso8601UTCTimeFormatter, self).__init__(log_format, time_format)

        self._TIMEZONE_UTC = tzutc()

    def formatTime(self, record, timeFormat=None):
        if timeFormat is not None:
            return super(Iso8601UTCTimeFormatter, self).formatTime(record, timeFormat)

        return str(datetime.datetime.fromtimestamp(record.created, self._TIMEZONE_UTC).isoformat(
            timespec='milliseconds')).replace('+00:00', 'Z')


class Log:

    def __init__(self, name):
        self.logger = self.configure_logger(name)

    def error(self, message):
        self.error_fmt({MESSAGE: message})

    def warn(self, message):
        self.warn_fmt({MESSAGE: message})

    def info(self, message):
        self.info_fmt({MESSAGE: message})

    def debug(self, message):
        self.debug_fmt({MESSAGE: message})

    def operation(self, oper: str, mesg: str, level: str = 'INFO', id: Any = None):
        log_kv = {'operation': oper, MESSAGE: mesg, 'id': id}
        if level == 'INFO':
            self.info_fmt(log_kv)
        if level == 'WARN':
            self.warn_fmt(log_kv)
        if level == 'ERROR':
            self.error_fmt(log_kv)

    def info_fmt(self, log_kv: dict, message: str = None):
        fmt = self._create_fmt(log_kv, message)
        self.logger.info(fmt)

    def warn_fmt(self, log_kv: dict, message: str = None):
        fmt = self._create_fmt(log_kv, message)
        self.logger.warning(fmt)

    def error_fmt(self, log_kv: dict, message: str = None):
        fmt = self._create_fmt(log_kv, message)
        self.logger.error(fmt)

    def debug_fmt(self, log_kv: dict, message: str = None):
        fmt = self._create_fmt(log_kv, message)
        self.logger.debug(fmt)

    def dump_command(self, command, status, response_text, url, body=None):
        self.info_fmt(
            {'operation': 'dump', 'command': command, 'status': status, 'response_text': response_text, 'url': url,
             'body': json.dumps(body)})

    def info_timer(self, method, path, object_name, time, num_of_calls=None, status=None, remote_address: str = None):
        self.info_fmt(
            {'system': 'icinga2', 'address': remote_address, 'method': method, 'path': path, 'object': object_name,
             'status': status, 'response_time': time})

    def _create_fmt(self, log_kv, message):
        if message:
            _log_kv = dict(log_kv)
            _log_kv[MESSAGE] = message
            fmt = self._format(_log_kv)
        else:
            fmt = self._format(log_kv)
        return fmt

    def _format(self, log_kv):
        fmt = ''
        sep = ''
        try:
            for k, v in log_kv.items():
                if v is not None:
                    if isinstance(v, numbers.Number):
                        fmt = f"{fmt}{sep}{k}={str(v)}"
                    elif ' ' in v or '"' in v:
                        if '"' in v:
                            j = v.replace('"', '\\"')
                            fmt = f"{fmt}{sep}{k}=\"{j}\""
                        else:
                            fmt = f"{fmt}{sep}{k}=\"{v}\""
                    else:
                        fmt = f"{fmt}{sep}{k}={v}"
                sep = ' '
            return fmt
        except Exception as err:
            return f"exception=\"{err}\" description=\"Parsing log entry\""

    def infotimer(self, system, method, path, time, num_of_calls=None, status=None):
        self.info_fmt({'system': system, 'method': method, 'path': path, 'status': status, 'response_time': time,
                       'calls': num_of_calls})

    def configure_logger(self, name):
        formatter = Iso8601UTCTimeFormatter('timestamp=%(asctime)s level=%(levelname)s %(message)s')

        logger = logging.getLogger(name)
        try:
            if conf.get('logger') and conf.get('logger').get('logfile'):
                log_filename = conf.get('logger').get('logfile')
                hdlr = logging.FileHandler(log_filename)
            else:
                hdlr = logging.StreamHandler(sys.stdout)

            if os.getenv('INDIS_LOG_FILE'):
                hdlr = logging.FileHandler(os.getenv('INDIS_LOG_FILE'))

            hdlr.setFormatter(formatter)
            logger.addHandler(hdlr)

            if conf.get('logger') and conf.get('logger').get('level'):
                log_level = conf.get('logger').get('level')
                logger.setLevel(log_level)

            if os.getenv('INDIS_LOG_LEVEL'):
                logger.setLevel(os.getenv('INDIS_LOG_LEVEL'))

        except Exception:

            hdlr = logging.StreamHandler(sys.stdout)
            if os.getenv('INDIS_LOG_FILE'):
                hdlr = logging.FileHandler(os.getenv('INDIS_LOG_FILE'))

            hdlr.setFormatter(formatter)
            logger.addHandler(hdlr)
            logger.setLevel('INFO')
            if os.getenv('INDIS_LOG_LEVEL'):
                logger.setLevel(os.getenv('INDIS_LOG_LEVEL'))

        return logger



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2021  Opsdis AB

    This file is part of indis - Icinga native directory importer service.

    indis is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    indis is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with indis.  If not, see <http://www.gnu.org/licenses/>.

"""
import json
import time
from typing import Dict, Set

import requests

from indis.cache import Cache
from indis.configuration import Configuration
from indis.logging import Log as log
from indis.model.service import HOST_SERVICE_SEPARATOR
from indis.output.output_writer import OutputWriter
from indis.provider.transfer import Transfer

UNKNOWN = 'unknown'
DELETED = 'deleted'
NOTMODIFIED = 'notmodified'
UPDATED = 'updated'
CREATED = 'created'
logger = log(__name__)


def map_write_status(status):
    if status == 201:
        return CREATED
    if status == 200:
        return UPDATED
    if status == 304:
        return NOTMODIFIED
    return UNKNOWN


def map_delete_status(status):
    if status == 200:
        return DELETED
    if status == 304:
        return NOTMODIFIED
    return UNKNOWN


class APIWriter(OutputWriter):

    def __init__(self, config: Configuration):
        super().__init__(config)
        self.cache = None
        self.stats_create = {}
        self.stats_create_template = {}
        self.stats_delete = {}
        self.con = Connection(self.config)
        self.transfer: Transfer = Transfer()

    def write(self, transfer: Transfer, cache: Cache):

        self.transfer = transfer
        # Cache the Transfer
        self.cache = cache

        # Write templates an objects in the correct dependency order
        for object_type in transfer.dependency_order():
            if transfer.__dict__[object_type]:
                # the transfer object is not empty
                # Create templates
                self.stats_create_template.update(self._write_template(object_type=object_type))
                # Create objects
                self.stats_create.update(self._write_object(object_type=object_type))

        logger.debug_fmt(log_kv=self.stats_create, message="api create/update template operations")

        logger.debug_fmt(log_kv=self.stats_create, message="api create/update object operations")

        # Handling delete objects
        for object_type in transfer.__dict__.keys():
            deleted = self.cache.deleted(object_type=object_type)

            self.stats_delete.update(self._delete_object(object_type=object_type, objects=deleted))

        logger.debug_fmt(log_kv=self.stats_delete, message="api delete object operations")

        if self.config.get('deploy'):
            self.deploy()

    def write_stats(self) -> Dict[str, Dict[str, Dict[str, int]]]:
        stats = dict()
        stats['object'] = {}
        for object_type in self.transfer.__dict__.keys():

            stats['object'][object_type] = {UPDATED: 0, CREATED: 0, DELETED: 0, NOTMODIFIED: 0, UNKNOWN: 0}
            if object_type in self.stats_create:
                for key, value in self.stats_create[object_type].items():
                    stats['object'][object_type][value] += 1
                for key, value in self.stats_delete[object_type].items():
                    stats['object'][object_type][value] += 1

        stats['template'] = {}
        for object_type in self.transfer.__dict__.keys():

            stats['template'][object_type] = {UPDATED: 0, CREATED: 0, DELETED: 0, NOTMODIFIED: 0, UNKNOWN: 0}
            if object_type in self.stats_create_template:
                for key, value in self.stats_create_template[object_type].items():
                    stats['template'][object_type][value] += 1
                for key, value in self.stats_delete[object_type].items():
                    stats['template'][object_type][value] += 1

        return stats

    def _write_template(self, object_type: str) -> Dict[str, Dict[str, str]]:
        stats = {object_type: dict()}

        for key, value in self.transfer.get_copy()[object_type].items():
            if value.object_type == 'template':
                status = self.con.create_object(object_name=value.object_name, object_type=object_type,
                                                body=value.to_json())
                stats[object_type][value.object_name] = map_write_status(status)

        return stats

    def _write_object(self, object_type: str) -> Dict[str, Dict[str, str]]:
        stats = {object_type: dict()}

        for key, value in self.transfer.get_copy()[object_type].items():
            if value.object_type == 'object':
                status = self.con.create_object(object_name=value.object_name, object_type=object_type,
                                                body=value.to_json())
                stats[object_type][value.object_name] = map_write_status(status)

        return stats

    def _delete_object(self, object_type: str, objects: Set[str]) -> Dict[str, Dict[str, str]]:
        stats = {object_type: dict()}

        for object_name in objects:
            status = self.con.delete_object(object_name=object_name, object_type=object_type)
            stats[object_type][object_name] = map_delete_status(status)
        return stats

    def deploy(self):
        self.con.deploy()


class Connection:

    def __init__(self, conf):
        user = conf.get('user')
        passwd = conf.get('password')
        self.url = conf.get('url')
        self.auth = (user, passwd)
        self.headers = {'Accept': 'application/json'}
        self.verify = False
        self.retries = 5

    def read_apply_rules(self):
        """
        Read apply rules
        TODO
        :return:
        """
        r = requests.get(f"{self.url}/serviceapplyrules",
                         auth=self.auth,
                         headers=self.headers,
                         verify=self.verify)
        # print(r)
        pass

    def create_object(self, object_name: str, object_type: str, body: str) -> int:

        type = self.get_url(object_type)

        url_postfix = ''
        # service special
        if type == "service":
            bd = json.loads(body)
            if 'host' in bd:
                host = bd['host']
                url_postfix = f"&host={host}"

        no_error = False
        start_time = time.time()
        status = None

        try:
            # Do PUT first - if the object exists it will be updated
            r = requests.put(f"{self.url}/{type}?name={object_name}{url_postfix}",
                             data=body,
                             auth=self.auth,
                             headers=self.headers,
                             verify=self.verify)

            if r.status_code == 404:
                # If the object do not exist do POST
                r = requests.post(f"{self.url}/{type}",
                                  data=body,
                                  auth=self.auth,
                                  headers=self.headers,
                                  verify=self.verify)

                status = r.status_code
                # no_error = True

            if r.status_code == 200 or r.status_code == 201:
                status = r.status_code

            if not no_error:
                r.raise_for_status()

            return r.status_code

        except requests.exceptions.HTTPError as err:
            logger.warn("put failed on: {} err {}".format(object_type, err))
            if r is None:
                raise err

            logger.dump_command('put', r.status_code, r.text, self.url, body)
            return r.status_code

        except requests.exceptions.RequestException as err:
            logger.error("put failed on: {} err {}".format(self.url, err))
            raise err
        finally:
            logger.info_timer('put', object_type, object_name, time.time() - start_time, status)

    def delete_object(self, object_name: str, object_type: str) -> int:
        type = self.get_url(object_type)
        url_postfix = ''
        # service special
        if type == "service":
            object_name_array = object_name.split(HOST_SERVICE_SEPARATOR)
            # object_name = object_name_array[1]
            url_postfix = f"&host={object_name_array[0]}"

        no_error = False
        start_time = time.time()
        status = None

        try:

            r = requests.delete(f"{self.url}/{type}?name={object_name}{url_postfix}",
                                auth=self.auth,
                                headers=self.headers,
                                verify=self.verify)

            if r.status_code == 200 or r.status_code == 201:
                status = r.status_code

            if not no_error:
                r.raise_for_status()

            return r.status_code

        except requests.exceptions.HTTPError as err:
            logger.warn("delete failed on: {} err {}".format(object_type, err))
            if r is None:
                raise err

            return r.status_code

        except requests.exceptions.RequestException as err:
            logger.error("delete failed on: {} err {}".format(self.url, err))
            raise err
        finally:
            logger.info_timer('delete', object_type, object_name, time.time() - start_time, status)

    def deploy(self) -> int:

        no_error = False
        start_time = time.time()
        status = None

        try:

            r = requests.post(f"{self.url}/config/deploy",
                              auth=self.auth,
                              headers=self.headers,
                              verify=self.verify)

            if r.status_code == 200 or r.status_code == 201:
                status = r.status_code

            if not no_error:
                r.raise_for_status()

            return r.status_code

        except requests.exceptions.HTTPError as err:
            logger.warn("get failed on: {} err {}".format('config/deployments', err))
            if r is None:
                raise err

            return r.status_code

        except requests.exceptions.RequestException as err:
            logger.error("get failed on: {} err {}".format(self.url, err))
            raise err
        finally:
            logger.info_timer('post', "config/deploy", "", time.time() - start_time, status)

    def get_url(self, object_type):
        type = ''
        if object_type == 'hosts':
            type = 'host'
        if object_type == 'hostgroups':
            type = 'hostgroup'
        if object_type == 'services':
            type = 'service'
        if object_type == 'servicegroups':
            type = 'servicegroup'
        if object_type == 'commands':
            type = 'command'
        if object_type == 'eventcommands':
            type = 'eventcommand'
        if object_type == 'notifications':
            type = 'notification'
        if object_type == 'timeperiods':
            type = 'timeperiod'
        if object_type == 'users':
            type = 'user'
        if object_type == 'usergroups':
            type = 'usergroup'
        if object_type == 'endpoints':
            type = 'endpoint'
        if object_type == 'zones':
            type = 'zone'
        return type



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2021  Opsdis AB

    This file is part of indis - Icinga native directory importer service.

    indis is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    indis is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with indis.  If not, see <http://www.gnu.org/licenses/>.

"""

from indis.model.basic_attributes import BasicAttributes

from indis.model.common import to_json, to_dict

HOST_SERVICE_SEPARATOR = '_'


class Service(BasicAttributes):
    __initialized = False

    def __init__(self, name: str, object_type: str = 'object', host_name: str = None):
        super().__init__(name=name, object_type=object_type)
        # 	Object name	Required. The host this service belongs to. There must be a Host object with that name.
        if object_type == 'object':
            self.host = host_name
            self.object_name = f"{self.host}{HOST_SERVICE_SEPARATOR}{name}"
        else:
            # Is a template
            self.object_name = f"{name}"
        self.__initialized = True
        self.display_name = name

    def __setattr__(self, name, value):
        if self.__initialized:
            self._ind.add(name)
            object.__setattr__(self, name, value)
        else:
            object.__setattr__(self, name, value)

    def to_json(self, padding: bool = False):
        if self.object_type == 'object':
            res = {'host': self.host}
            return to_json(self, res)
        else:
            return to_json(self)

    def to_dict(self):
        if self.object_type == 'object':
            res = {'host': self.host}
            return to_dict(self, res)
        else:
            return to_dict(self)



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2021  Opsdis AB

    This file is part of indis - Icinga native directory importer service.

    indis is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    indis is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with indis.  If not, see <http://www.gnu.org/licenses/>.

"""

import json
from abc import abstractmethod
from typing import Dict, Any


class Common:

    def __init__(self, name, object_type: str):
        self.object_name = name
        self.object_type = object_type

    @abstractmethod
    def to_json(self) -> str:
        pass

    @abstractmethod
    def to_dict(self) -> dict:
        pass


def to_json(obj, initial=None, padding: bool = False) -> str:
    res = to_dict(obj, initial, padding)
    return json.dumps(res)


def to_dict(obj=None, initial=None, padding: bool = False) -> Dict[str, Any]:
    res = dict()
    for key, value in obj.__dict__.items():
        if key.startswith('_'):
            # Remove all protected attributes
            continue
        if key in obj._ind:
            res[key] = value
        elif isinstance(obj.__dict__[key], dict) and obj.__dict__[key]:
            res[key] = value
        elif isinstance(obj.__dict__[key], list) and obj.__dict__[key]:
            res[key] = value
        elif padding:
            res[key] = None
    if initial is not None:
        res.update(initial)
    return res



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2021  Opsdis AB

    This file is part of indis - Icinga native directory importer service.

    indis is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    indis is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with indis.  If not, see <http://www.gnu.org/licenses/>.

"""

from indis.model.basic_attributes import BasicAttributes
from indis.model.common import to_json, to_dict


class Host(BasicAttributes):
    __initialized = False

    def __init__(self, name, object_type: str = 'object'):
        super().__init__(name=name, object_type=object_type)

        # String Optional. The host’s IPv4 address. Available as command runtime macro $address$ if set.
        self.address = None
        # String	Optional. The host’s IPv6 address. Available as command runtime macro $address6$ if set.
        self.address6 = None
        self.__initialized = True

    def __setattr__(self, name, value):
        if self.__initialized:
            self._ind.add(name)
            object.__setattr__(self, name, value)
        else:
            object.__setattr__(self, name, value)

    def to_json(self):

        return to_json(self)

    def to_dict(self):
        return to_dict(self)



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2021  Opsdis AB

    This file is part of indis - Icinga native directory importer service.

    indis is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    indis is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with indis.  If not, see <http://www.gnu.org/licenses/>.

"""

from abc import abstractmethod

from indis.configuration import Configuration
from indis.provider.source_reader import SourceReader
from indis.provider.transfer import Transfer


class Source:

    def __init__(self, config: Configuration, reader: SourceReader):
        self.config = config
        self.reader = reader

    @abstractmethod
    def fetch(self) -> Transfer:
        pass



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2021  Opsdis AB

    This file is part of indis - Icinga native directory importer service.

    indis is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    indis is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with indis.  If not, see <http://www.gnu.org/licenses/>.

"""
import inspect
import logging

from indis.provider.source import Source
from indis.provider.transfer import Transfer

cmdblogging = logging.getLogger(__name__)


class Factory:

    def __init__(self, cmdb, configuration, reader: None):

        self.cmdb = cmdb
        try:
            self.source_name = configuration.get('source')[self.cmdb].get('name')
            self.source_module = configuration.get('source')[self.cmdb].get('class')
            self.source_interface = configuration.get('source')[self.cmdb].get('access').get('reader_inf')
            if not self.source_name or not self.source_module or not self.source_interface:
                raise AttributeError('Missing mandatory config for source')
        except Exception as err:
            raise AttributeError(err)

        self.cmdb_reader_name = None if configuration.get('source')[self.cmdb].get('access').get('reader') is None \
            else configuration.get('source')[cmdb].get('access').get('reader')

        self.configuration = configuration

        self.reader = None
        if reader:
            if isinstance(reader, Factory._get_class(self.source_interface)):
                self.reader = reader
            else:
                raise AttributeError("The reader object is not according to interface")

        # create Source object

        clazz = Factory._get_class(self.source_module)
        self.source_obj: Source = clazz(self.configuration.get(self.configuration.get('source')[self.cmdb].get('name')),
                                        self._get_reader())
        if not isinstance(self.source_obj, Source):
            raise AttributeError("The reader object is not according to interface")

    def _get_reader(self):
        """
        The method find and instantiate the Reader class for the cmdb.
        It support Reader __init__ constructors with 1 or 2 arguments.
        If 1 argument -> configuration
        If 2 arguments -> configuration, cmdb_name
        If using 2 arguments the Reader class do not need to know the configuration
        name for the cmdb.
        :return: object of the Reader class
        """

        if self.reader:
            return self.reader
        elif self.configuration.get('source')[self.cmdb].get('access').get('reader'):
            clazz = Factory._get_class(self.configuration.get('source')[self.cmdb].get('access').get('reader'))

            arg_list = inspect.getfullargspec(clazz.__init__).args

            if clazz and len(arg_list) == 2:
                return clazz(self.configuration.get(self.configuration.get('source')[self.cmdb].get('name')))
            else:
                return None
        else:
            return None

    def fetch(self) -> Transfer:
        """
        The method return a method, fetch, that is specific for the CMDB provider.
        The fetch method that is return must by itself return a dictionary of Host objects with the key of the host name
        and the hostgroup set with the name of the host groups as plain string.
        :return:
        """

        transfer = self.source_obj.fetch()

        cmdblogging.info("Using source factory {} ".format(self.source_name))

        return transfer

    @staticmethod
    def _get_module(name):
        return __import__(name, fromlist=[''])

    @staticmethod
    def _get_class(kls):

        try:
            parts = kls.split('.')
            module = ".".join(parts[:-1])
            m = __import__(module)
            for comp in parts[1:]:
                m = getattr(m, comp)
            return m
        except Exception as err:
            cmdblogging.error(f"Class {kls} failed {err}")
            raise err



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2021  Opsdis AB

    This file is part of indis - Icinga native directory importer service.

    indis is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    indis is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with indis.  If not, see <http://www.gnu.org/licenses/>.

"""
from typing import Dict

from indis.cache import Cache
from indis.configuration import Configuration
from indis.output.output_writer import OutputWriter
from indis.provider.transfer import Transfer


class JsonFileWriter(OutputWriter):

    def __init__(self, config: Configuration):
        super().__init__(config)
        self.stats = dict()

    def write(self, transfer: Transfer, cache: Cache):

        for object_type in transfer.get_keys():
            if self.config and self.config.get('directory'):
                with open(self.config.get('directory') + "/" + object_type + '.json', 'w') as fd:
                    fd.write('[')
                    sep = ''
                    self.stats[object_type] = {'created': 0}
                    for key, value in transfer.get_copy()[object_type].items():
                        self.stats[object_type]['created'] += 1
                        fd.write(sep)
                        fd.write(value.to_json())
                        sep = ','
                    fd.write(']')
            else:
                print('[', end='')
                sep = ''
                self.stats[object_type] = {'created': 0}
                for key, value in transfer.get_copy()[object_type].items():
                    self.stats[object_type]['created'] += 1
                    print(sep, end='')
                    print(value.to_json(), end='')
                    sep = ','
                print(']', end='')

    def write_stats(self) -> Dict[str, Dict[str, Dict[str, int]]]:
        return self.stats



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2021  Opsdis AB

    This file is part of indis - Icinga native directory importer service.

    indis is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    indis is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with indis.  If not, see <http://www.gnu.org/licenses/>.

"""
from typing import Dict

from indis.configuration import Configuration
from indis.processor.common import Processing
# The import of the classes that is subclasses of Processing
from indis.processor.groups import Groups
from indis.processor.vars import Vars
from indis.provider.transfer import Transfer


def processing(transfer: Transfer, config: Configuration) -> Dict[str, int]:
    """
    Find all processing "plugins", instantiate them and and execute the objects process() method.
    All classes to execute must be part of above imports
    Must been imported
    :param transfer:
    :param config:
    :return:
    """
    processed = {}
    for clazz in Processing.__subclasses__():
        proc_obj: Processing = clazz(transfer=transfer, config=config)
        processed[type(proc_obj).__name__] = proc_obj.process()
    return processed



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2021  Opsdis AB

    This file is part of indis - Icinga native directory importer service.

    indis is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    indis is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with indis.  If not, see <http://www.gnu.org/licenses/>.

"""

from typing import List, Dict, Any

from indis.model.common import Common


class BasicAttributes(Common):
    __initialized = False

    def __init__(self, name: str, object_type: str):
        super().__init__(name=name, object_type=object_type)
        # self.object_name = name
        # String	Optional. A short description of the service.
        self.display_name: str = ''
        # Array of object names	Optional. The service groups this service belongs to.
        self.groups: List[str] = list()
        # Dictionary	Optional. A dictionary containing custom variables that are specific to this service.
        self.vars: Dict[str, Any] = dict()
        # Object name	Required. The name of the check command.
        self.check_command: str = ''
        # Number	Optional. The number of times a service is re-checked before changing into a hard state. Defaults to 3.
        self.max_check_attempts: int = 3
        # Object name	Optional. The name of a time period which determines when this service should be checked. Not set by default (effectively 24x7).
        self.check_period: str = ''
        # Duration	Optional. Check command timeout in seconds. Overrides the CheckCommand’s timeout attribute.
        self.check_timeout: int = 0
        # Duration	Optional. The check interval (in seconds). This interval is used for checks when the service is in a HARD state. Defaults to 5m.
        self.check_interval: str = '5m'
        # Duration	Optional. The retry interval (in seconds). This interval is used for checks when the service is in a SOFT state. Defaults to 1m. Note: This does not affect the scheduling after a passive check result.
        self.retry_interval: int = 0
        # Boolean	Optional. Whether notifications are enabled. Defaults to true.
        self.enable_notifications: bool = True
        # Boolean	Optional. Whether active checks are enabled. Defaults to true.
        self.enable_active_checks: bool = True
        # Boolean	Optional. Whether passive checks are enabled. Defaults to true.
        self.enable_passive_checks: bool = True
        # Boolean	Optional. Enables event handlers for this host. Defaults to true.
        self.enable_event_handler: bool = True
        # Boolean	Optional. Whether flap detection is enabled. Defaults to false.
        self.enable_flapping: bool = False
        # Boolean	Optional. Whether performance data processing is enabled. Defaults to true.
        self.enable_perfdata: bool = True
        # Object name	Optional. The name of an event command that should be executed every time the service’s state changes or the service is in a SOFT state.
        self.event_command: str = ''
        # Number	Optional. Flapping upper bound in percent for a service to be considered flapping. 30.0
        self.flapping_threshold_high: float = 30.0
        # Number	Optional. Flapping lower bound in percent for a service to be considered not flapping. 25.0
        self.flapping_threshold_low: float = 25.0
        # Array	Optional. A list of states that should be ignored during flapping calculation. By default no state is ignored.
        self.flapping_ignore_states: List[str] = list()
        # Boolean	Optional. Treat all state changes as HARD changes. See here for details. Defaults to false.
        self.volatile: bool = False
        # Object name	Optional. The zone this object is a member of. Please read the distributed monitoring chapter for details.
        self.zone: str = ''
        # Object name	Optional. The endpoint where commands are executed on.
        self.command_endpoint: str = ''
        # String	Optional. Notes for the service.
        self.notes: str = ''
        # String	Optional. URL for notes for the service (for example, in notification commands).
        self.notes_url: str = ''
        # String	Optional. URL for actions for the service (for example, an external graphing tool).
        self.action_url: str = ''
        # String	Optional. Icon image for the service. Used by external interfaces only.
        self.icon_image: str = ''
        # String	Optional. Icon image description for the service. Used by external interface only.
        self.icon_image_alt: str = ''

        self.imports: List[str] = list()
        self._ind = {'object_name', 'object_type'}
        self.__initialized = True

    def __setattr__(self, name, value):
        if self.__initialized:
            self._ind.add(name)
            object.__setattr__(self, name, value)
        else:
            object.__setattr__(self, name, value)

    def __getattr__(self, name):
        if name == 'groups':
            return list(set(self.groups))
        else:
            return object.__getattr__(self, name)



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2021  Opsdis AB

    This file is part of indis - Icinga native directory importer service.

    indis is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    indis is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with indis.  If not, see <http://www.gnu.org/licenses/>.

"""

import hashlib
from abc import abstractmethod
from typing import Set, Dict, Any

import redis

from indis.configuration import Configuration
from indis.provider.transfer import Transfer


class Cache:

    def __init__(self, transfer: Transfer):
        self.transfer = transfer

    @abstractmethod
    def diff(self) -> Transfer:
        pass

    @abstractmethod
    def deleted(self, object_type: str) -> Set[str]:
        pass

    @abstractmethod
    def updated(self, object_type: str) -> Set[str]:
        pass

    @abstractmethod
    def created(self, object_type: str) -> Set[str]:
        pass

    @abstractmethod
    def rollback(self):
        pass


class NoCache(Cache):

    def __init__(self, transfer: Transfer):
        super().__init__(transfer)

    def diff(self) -> Transfer:
        return None

    def deleted(self, object_type: str) -> Set[str]:
        return set()

    def updated(self, object_type: str) -> Set[str]:
        return set()

    def created(self, object_type: str) -> Set[str]:
        return set()

    def rollback(self):
        pass


class RedisCache(Cache):

    def __init__(self, prefix: str, transfer: Transfer, config: Configuration):
        super().__init__(transfer=transfer)
        self.redis_host = 'localhost' if config.get('host') is None else config.get('host')
        self.redis_port = '6379' if config.get('port') is None else config.get('port')
        self.redis_db = '0' if config.get('db') is None else config.get('db')
        self.redis_auth = None if config.get('auth') is None else config.get('auth')
        self.key_prefix = f"{prefix}:"
        self.con = self._get_cache_connection()

        self.transfer_existing: Dict[str, Dict[str, Any]] = dict()
        for object_type, value in transfer.__dict__.items():
            self.transfer_existing[object_type] = dict()
            for object in self.con.smembers(f"{self.key_prefix}{object_type}"):
                self.transfer_existing[object_type][object] = self.con.get(f"{self.key_prefix}{object_type}:{object}")
                self.con.delete(f"{self.key_prefix}{object_type}:{object}")
            self.con.delete(f"{self.key_prefix}{object_type}")

        self.transfer_current: Dict[str, Dict[str, Any]] = dict()
        pipe = self.con.pipeline()
        for object_type, value in transfer.__dict__.items():
            self.transfer_current[object_type] = dict()
            for key, obj in value.items():
                pipe.sadd(f"{self.key_prefix}{object_type}", key)
                sha = hashlib.sha256(obj.to_json().encode())
                self.transfer_current[object_type][key] = sha.hexdigest()
                pipe.set(f"{self.key_prefix}{object_type}:{key}", sha.hexdigest())
        pipe.execute()

    def rollback(self):
        """
        replace the current with existing
        :return:
        """
        pipe = self.con.pipeline()
        for object_type, value in self.transfer_existing.items():
            pipe.delete(f"{self.key_prefix}{object_type}")
            for key, sha_value in value.items():
                pipe.sadd(f"{self.key_prefix}{object_type}", key)
                pipe.set(f"{self.key_prefix}{object_type}:{object}", sha_value)

        pipe.execute()

    def diff(self) -> Transfer:
        pass

    def deleted(self, object_type: str) -> Set[str]:
        return self.transfer_existing[object_type].keys() - self.transfer_current[object_type].keys()

    def updated(self, object_type: str) -> Set[str]:
        return set()

    def created(self, object_type: str) -> Set[str]:
        return set()

    def _get_cache_connection(self):
        return redis.Redis(host=self.redis_host, port=self.redis_port, db=self.redis_db, password=self.redis_auth,
                           decode_responses=True)



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2021  Opsdis AB

    This file is part of indis - Icinga native directory importer service.

    indis is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    indis is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with indis.  If not, see <http://www.gnu.org/licenses/>.

"""

from typing import Dict, Any, Set

from indis.model.command import Command
from indis.model.endpoint import EndPoint
from indis.model.group import Group
from indis.model.host import Host
from indis.model.host_dependency import HostDependency
from indis.model.notification import Notification
from indis.model.service import Service
from indis.model.service_dependency import ServiceDependency
from indis.model.timeperiod import TimePeriod
from indis.model.user import User
from indis.model.usergroup import UserGroup
from indis.model.zone import Zone


class Transfer:
    # Value object passing data
    def __init__(self):

        # dict hostname-> Host
        self.hosts: Dict[str, Host] = {}
        self.services: Dict[str, Service] = {}

        self.hostgroups: Dict[str, Group] = {}
        self.servicegroups: Dict[str, Group] = {}

        self.host_dependencies: Dict[str, HostDependency] = {}
        self.service_dependencies: Dict[str, ServiceDependency] = {}

        self.notifications: Dict[str, Notification] = {}
        self.commands: Dict[str, Command] = {}
        self.eventcommands: Dict[str, Command] = {}

        self.timeperiods: Dict[str, TimePeriod] = {}
        self.users: Dict[str, User] = {}
        self.usergroups: Dict[str, UserGroup] = {}

        self.endpoints: Dict[str, EndPoint] = {}
        self.zones: Dict[str, Zone] = {}

    # Set of host name that should be excluded
    # self.hosts_exclude: Dict[str, Any] = {}

    def __setattr__(self, key, value):

        if not isinstance(value, dict):
            raise ValueError(f"{key} must be a dict")
        else:
            super().__setattr__(key, value)

    def __getattr__(self, key):

        if key in self.__dict__:
            return self.__dict__[key]
        else:
            raise AttributeError(f'{self.__class__.__name__} has no attribute name {key}')

    def get_copy(self) -> Dict[str, Any]:
        set_dict = {}
        for k, v in self.__dict__.items():
            if k and v:
                set_dict[k] = v
        return set_dict

    def get_keys(self) -> Set[str]:
        set_dict = set()
        for k, v in self.__dict__.items():
            if k and v:
                set_dict.add(k)
        return set_dict

    def get_by_name(self, name) -> Dict[str, Any]:
        return self.__dict__[name]

    def stats(self) -> Dict[str, int]:
        stats = {}
        for key, value in self.__dict__.items():
            stats[key] = len(value)
        return stats

    def dependency_order(self):
        return ['endpoints', 'zones', 'hostgroups', 'commands', 'eventcommands', 'usergroups', 'users', 'notifications', 'timeperiods',
                'servicegroups', 'hosts', 'services', 'host_dependencies', 'service_dependencies']



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2021  Opsdis AB

    This file is part of indis - Icinga native directory importer service.

    indis is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    indis is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with indis.  If not, see <http://www.gnu.org/licenses/>.

"""

from typing import List, Dict, Any

from indis.model.common import Common, to_json, to_dict


class Command(Common):
    """
    Command is the same for CheckCommand, NotificationCommand and EventCommand
    """
    __initialized = False

    def __init__(self, name: str, object_type: str = 'object'):
        super().__init__(name=name, object_type=object_type)
        # self.object_name = name
        # Array	Required. The command.: This can either be an array of individual command arguments. Alternatively a
        # string can be specified in which case the shell interpreter (usually /bin/sh) takes care of parsing the
        # command.
        # When using the “arguments” attribute this must be an array. Can be specified as function for advanced implementations.
        self.command: List[str] = list()
        # Dictionary	Optional. A dictionary of macros which should be exported as environment variables prior to executing the command.
        self.env: Dict[str, str] = dict()
        # Dictionary	Optional. A dictionary containing custom variables that are specific to this command.
        self.vars: Dict[str, Any] = dict()
        # Duration	Optional. The command timeout in seconds. Defaults to 1m.
        self.timeout: str = ''
        # Dictionary	Optional. A dictionary of command arguments.
        self.arguments: Dict[str, str] = dict()

        self._ind = {'object_name', 'object_type'}
        self.__initialized = True

    def __setattr__(self, name, value):
        if self.__initialized:

            self._ind.add(name)
            object.__setattr__(self, name, value)
        else:
            object.__setattr__(self, name, value)

    def to_json(self):
        return to_json(self)

    def to_dict(self):
        return to_dict(self)



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2021  Opsdis AB

    This file is part of indis - Icinga native directory importer service.

    indis is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    indis is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with indis.  If not, see <http://www.gnu.org/licenses/>.

"""

from indis.cache.cache import Cache, RedisCache, NoCache
from indis.configuration import Configuration
from indis.logging import Log as log
from indis.provider.transfer import Transfer

logger = log(__name__)


def factory(prefix: str, config: Configuration, transfer: Transfer) -> Cache:
    if config:
        if config.get('redis'):
            logger.info_fmt({'cache': 'redis'}, 'Use cache')
            return RedisCache(prefix=prefix, transfer=transfer, config=config.get('redis'))
        else:
            logger.info_fmt({'cache': 'NoCache'}, 'Use cache')
            return NoCache(transfer)
    else:
        logger.info_fmt({'cache': 'NoCache'}, 'Use cache')
        return NoCache(transfer)



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2021  Opsdis AB

    This file is part of indis - Icinga native directory importer service.

    indis is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    indis is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with indis.  If not, see <http://www.gnu.org/licenses/>.

"""

from indis.processor.common import Processing


class Vars(Processing):

    def process(self) -> int:
        pc = self.config
        count = 0
        if pc and 'vars' in pc:

            for object_type in self.transfer.get_keys():
                if 'common' in pc.get('vars'):
                    for key, value in pc.get('vars').get('common').items():
                        for object_name in self.transfer.hosts.keys():
                            self.transfer.hosts[object_name].vars[key] = value
                            count += 1

                if object_type in pc.get('vars'):
                    # For each object specific
                    for key, value in pc.get('vars').get(object_type).items():
                        # for object_name in self.transfer.hosts.keys():
                        for object_name in self.transfer.get_by_name('hosts').keys():
                            # self.transfer.hosts[object_name].vars[key] = value
                            self.transfer.get_by_name('hosts')[object_name].vars[key] = value
                            count += 1
        return count



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2021  Opsdis AB

    This file is part of indis - Icinga native directory importer service.

    indis is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    indis is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with indis.  If not, see <http://www.gnu.org/licenses/>.

"""

from typing import Dict, Any, List

from indis.model.common import Common, to_json, to_dict


class Notification(Common):
    __initialized = False

    def __init__(self, name: str, object_type: str = 'object'):
        super().__init__(name=name, object_type=object_type)
        # self.object_name: str = name
        # Object name	Required. The name of the host this notification belongs to.
        self.host_name: str = ''
        # Object name	Optional. The short name of the service this notification belongs to. If omitted, this notification object is treated as host notification.
        self.service_name: str = ''
        # Dictionary	Optional. A dictionary containing custom variables that are specific to this notification object.
        self.vars: Dict[str, Any] = dict()
        # Array of object names	Required. A list of user names who should be notified. Optional. if the user_groups attribute is set.
        self.users: List[str] = list()
        # Array of object names	Required. A list of user group names who should be notified. Optional. if the users attribute is set.
        self.user_groups: List[str] = list()
        # Dictionary	Optional. A dictionary containing begin and end attributes for the notification.
        self.times: Dict[str, str] = dict()
        # Object name	Required. The name of the notification command which should be executed when the notification is triggered.
        self.command: str = ''
        # Duration	Optional. The notification interval (in seconds). This interval is used for active notifications. Defaults to 30 minutes. If set to 0, re-notifications are disabled.
        self.interval: int = 0
        # Object name	Optional. The name of a time period which determines when this notification should be triggered. Not set by default (effectively 24x7).
        self.period: str = ''
        # Object name	Optional. The zone this object is a member of. Please read the distributed monitoring chapter for details.
        self.zone: str = ''
        # Array	Optional. A list of type filters when this notification should be triggered. By default everything is matched.
        self.types: List[str] = list()
        # Array	Optional. A list of state filters when this notification should be triggered. By default everything is matched. Note that the states filter is ignored for notifications of type Acknowledgement!
        self.states: List[str] = list()

        self._ind = {'object_name', 'object_type'}
        self.__initialized = True

    def __setattr__(self, name, value):
        if self.__initialized:

            self._ind.add(name)
            object.__setattr__(self, name, value)
        else:
            object.__setattr__(self, name, value)

    def to_json(self):
        return to_json(self)

    def to_dict(self):
        return to_dict(self)



import unittest

from indis.model.host import Host
from indis.model.host_dependency import HostDependency
from indis.model.service import Service
from indis.model.group import Group
from indis.model.dependency import Dependency
from indis.model.service_dependency import ServiceDependency
from indis.model.zone import Zone


class TestBasicAttributes(unittest.TestCase):

    def test_host(self) -> None:
        print('Host')
        host = Host("kalle")
        host.vars["sune"] = "sune_value"
        host.vars["kalle"] = {"level": "two"}
        host.vars["alist"] = ["1", "2"]
        host.address = "172.25.1.111"
        host.groups.append("group1")

        print(host.to_json())
        self.assertTrue(not False)

    def test_service(self) -> None:
        print('Service')
        service = Service("aservice", "kalle")
        service.vars["alist"] = ["1", "2"]
        print(service.to_json())

        self.assertTrue(not False)

    def test_group(self) -> None:
        print('Group')
        group = Group("group1")
        group.display_name = "display_group1"
        group.groups.append("asubgroup")
        print(group.to_json())

        self.assertTrue(not False)

    def test_dependency(self) -> None:
        print('Dependency')
        dep_host = Dependency(name="mydep", parent_host="host1", child_host="host2")
        dep_host.states.append('Up')
        print(dep_host.to_json())

        self.assertTrue(not False)

    def test_host_dependency(self) -> None:
        print('HostDependency')
        dep_host = HostDependency(name="myhostdep", parent_host="host1", child_host="host2")
        dep_host.states.append('Up')
        print(dep_host.to_json())

        self.assertTrue(not False)

    def test_service_dependency(self) -> None:
        print('ServiceDependency')
        dep_service = ServiceDependency(name="myservicedep", parent_host="host1", child_host="host2", parent_service='foo', child_service='bar')
        dep_service.states.append('Down')
        print(dep_service.to_json())

        self.assertTrue(not False)

    def test_zone(self) -> None:
        print('zone')
        zone = Zone(name='foo')
        zone.is_global = True
        print(zone.to_json())
        print(zone.to_dict())

        self.assertTrue(not False)


if __name__ == '__main__':
    unittest.main()



from os.path import dirname, join

from setuptools import setup, find_packages


def read(fname):
    return open(join(dirname(__file__), fname)).read()


setup(
    name='indis',
    version_config={
        "template": "{tag}",
        "dev_template": "{tag}.dev{ccount}",
        "dirty_template": "{tag}.post{ccount}+git.{sha}.dirty",
        "starting_version": "0.0.1",
        "version_callback": None,
        "version_file": None,
        "count_commits_from_version_file": False,
        "branch_formatter": None
    },
    setup_requires=['setuptools-git-versioning'],
    packages=find_packages(exclude=['demo']),
    author='thenodon',
    author_email='ahaal@redbridge.se',
    url='https://github.com/opsdis/indis',
    license='GPLv3',
    include_package_data=True,
    zip_safe=False,
    description='indis - Icinga native director import service',
    install_requires=read('requirements.txt').split(),
    python_requires='>=3.8',
)



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2021  Opsdis AB

    This file is part of indis - Icinga native directory importer service.

    indis is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    indis is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with indis.  If not, see <http://www.gnu.org/licenses/>.

"""

from typing import List

from indis.model.common import Common, to_json, to_dict


class UserGroup(Common):
    """
    Command is the same for CheckCommand, NotificationCommand and EventCommand
    """
    __initialized = False

    def __init__(self, name: str, object_type: str = 'object'):
        super().__init__(name=name, object_type=object_type)
        # String	Optional. A short description of the user group.
        self.display_name: str = ''
        # Array of object names	Optional. An array of nested group names.
        self.groups: List[str] = list()

        self._ind = {'object_name', 'object_type'}
        self.__initialized = True

    def __setattr__(self, name, value):
        if self.__initialized:

            self._ind.add(name)
            object.__setattr__(self, name, value)
        else:
            object.__setattr__(self, name, value)

    def to_json(self):
        return to_json(self)

    def to_dict(self):
        return to_dict(self)



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2021  Opsdis AB

    This file is part of indis - Icinga native directory importer service.

    indis is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    indis is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with indis.  If not, see <http://www.gnu.org/licenses/>.

"""

from indis.model.common import Common, to_json, to_dict


class Dependency(Common):
    __initialized = False

    def __init__(self, name: str, parent_host: str, child_host: str):
        super().__init__(name=name, object_type='object')
        # self.object_name = name
        # Object name	Required. The parent host.
        self.parent_host_name = parent_host
        # Object name	Optional. The parent service. If omitted, this dependency object is treated as host dependency.
        self.parent_service_name = ''
        # Object name	Required. The child host.
        self.child_host_name = child_host
        # Object name	Optional. The child service. If omitted, this dependency object is treated as host dependency.
        self.child_service_name = ''
        # Boolean	Optional. Whether to disable checks (i.e., don’t schedule active checks and drop passive results) when this dependency fails. Defaults to false.
        self.disable_checks = False
        # Boolean	Optional. Whether to disable notifications when this dependency fails. Defaults to true.
        self.disable_notifications = True
        # Boolean	Optional. Whether to ignore soft states for the reachability calculation. Defaults to true.
        self.ignore_soft_states = True
        # Object name	Optional. Time period object during which this dependency is enabled.
        self.period = ''
        # Array	Optional. A list of state filters when this dependency should be OK. Defaults to [ OK, Warning ] for services and [ Up ] for hosts.
        self.states = list()

        self._ind = {'object_name', 'object_type', 'parent_host_name', 'child_host_name'}
        self.__initialized = True

    def __setattr__(self, name, value):
        if self.__initialized:

            self._ind.add(name)
            object.__setattr__(self, name, value)
        else:
            object.__setattr__(self, name, value)

    def to_json(self):
        return to_json(self)

    def to_dict(self):
        return to_dict(self)



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2021  Opsdis AB

    This file is part of indis - Icinga native directory importer service.

    indis is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    indis is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with indis.  If not, see <http://www.gnu.org/licenses/>.

"""

import logging
from typing import Dict

from indis.cache import Cache
from indis.output.output_writer import OutputWriter
from indis.provider.transfer import Transfer

cmdblogging = logging.getLogger(__name__)


class Factory:

    def __init__(self, cmdb, configuration):

        self.cmdb = cmdb

        self.output_name = configuration.get('output').get('writer')
        if not self.output_name:
            self.output_name = "indis.output.json_writer.JsonFileWriter"

        self.configuration = configuration.get('output').get('configuration')

        self.reader = None

        clazz = Factory._get_class(self.output_name)
        self.writer_obj: OutputWriter = clazz(self.configuration)
        if not isinstance(self.writer_obj, OutputWriter):
            raise AttributeError("The writer object is not according to interface")

    def write(self, transfer: Transfer, cache: Cache):
        """
        The method return a method, fetch, that is specific for the CMDB provider.
        The fetch method that is return must by itself return a dictionary of Host objects with the key of the host name
        and the hostgroup set with the name of the host groups as plain string.
        :return:
        """

        cmdblogging.info("Using writer factory {} ".format(self.writer_obj))
        self.writer_obj.write(transfer, cache)

    def write_stats(self) -> Dict[str, Dict[str, int]]:
        return self.writer_obj.write_stats()

    @staticmethod
    def _get_module(name):
        return __import__(name, fromlist=[''])

    @staticmethod
    def _get_class(kls):

        try:
            parts = kls.split('.')
            module = ".".join(parts[:-1])
            m = __import__(module)
            for comp in parts[1:]:
                m = getattr(m, comp)
            return m
        except Exception as err:
            cmdblogging.error(f"Class {kls} failed {err}")
            raise err



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2021  Opsdis AB

    This file is part of indis - Icinga native directory importer service.

    indis is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    indis is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with indis.  If not, see <http://www.gnu.org/licenses/>.

"""

import argparse
import json
import os
import traceback
# from cmdb2monitor.cmdbmo import create_monitor
from distutils.util import strtobool
from typing import Dict, Tuple

import indis.configuration as conf
from indis.cache import factory as cache_factory
from indis.configuration import Configuration
from indis.logging import Log as log
from indis.output.api_writer import Connection
from indis.output_factory import Factory as output_factory
from indis.processor import processing
from indis.source_factory import Factory as source_factory

logger = log(__name__)


def execute(source_name, dryrun: bool, source_reader=None) -> Tuple[Dict[str, int], Dict[str, Dict[str, int]]]:
    if not source_name and not os.getenv('INDIS_PROVIDER'):
        raise Exception("provider name not set")
    source_name = os.getenv('INDIS_PROVIDER', source_name)
    dryrun = bool(strtobool(os.getenv('INDIS_DRYRUN', str(dryrun))))

    try:
        # Get the output class
        output = output_factory(source_name, conf.Configuration)
        logger.info_fmt({'name': output.output_name}, f"created output factory")
        # Get the source provider class and
        source = source_factory(source_name, conf.Configuration, source_reader)
        logger.info_fmt({'name': source.source_module}, f"created source factory")
        logger.info_fmt({'name': source.source_interface}, f"created source factory reader")

        # Fetch transfer from source
        transfer = source.fetch()
        logger.info_fmt(transfer.stats(), "created objects")
        # Process
        processed = processing(transfer=transfer, config=conf.Configuration.get('processor'))
        logger.info_fmt(processed, f"processed")

        # Get cache
        cache_imp = cache_factory(prefix=source_name, config=conf.Configuration.get('cache'), transfer=transfer)

        # Apply rules - TODO
        con = Connection(conf.Configuration.get('output').get('configuration'))
        con.read_apply_rules()

        # Write output
        output.write(transfer=transfer, cache=cache_imp)
        logger.info("output executed")
        return transfer.stats(), output.write_stats()

    except Exception as err:
        raise err


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='indis - the icinga2 native director importer service')

    parser.add_argument('-d', '--dryrun', action="store_true",
                        dest="dry_run", help="Dry run no updates")

    parser.add_argument('-f', '--configfile',
                        dest="configfile", help="configuration file")

    parser.add_argument('-s', '--source',
                        dest="source_name", help="source provider to run")

    parser.add_argument('-v', '--verbose', action="store_true",
                        dest="verbose", help="verbose output of processing")

    args = parser.parse_args()

    if args.configfile:
        Configuration(args.configfile)
    else:
        Configuration()

    if args.dry_run:
        dry_run = True
    else:
        dry_run = False

    try:
        transfer_stats, ops_stat = execute(source_name=args.source_name, dryrun=dry_run)
        if args.verbose:
            verbose_output = {'source': transfer_stats, 'processed': ops_stat}
            print(json.dumps(verbose_output, indent=4, sort_keys=True))

    except Exception as err:
        traceback.print_exc()
        print(err)
